(*
  Autor" Alojz Bendak 2006
*)
unit figureobject;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, ExtCtrls, Buttons, ComCtrls, figureutils;

type
  TFigureObject = class (TObject)
  private
    Modified: boolean;
    FParts: TParts;
    FLeft: Integer;
    FTop: Integer;
    FColor: TColor;
    FFigureObjectType: TFigureObjectType;
    FCanvas: TCanvas;
    FBorderColor: TColor;
    FCanGoDown: boolean;
    FHeight: Integer;
    FWidth: Integer;
    FFigureRotated: TRotatedType;
    procedure WriteIntoGrid(DoMark: boolean);
    procedure SetRotatedProperties(RValue: TFigureObjectType; RRotated: TRotatedType;
      var RParts: TParts; var RWidth, RHeight: integer);
    procedure SetFigureType(Value: TFigureObjectType; Rotated: TRotatedType);
    procedure SetColor(const Value: TColor);
    procedure SetFigureObjectType(const Value: TFigureObjectType);
    procedure SetBorderColor(const Value: TColor);
    function GetCanGoDown: boolean;
    procedure SetFigureRotated(const Value: TRotatedType);
    function GetCanGoToLeft: boolean;
    function GetCanGoToRight: boolean;
    function GetEmpty: boolean;
    function GetNextRotatedDirection(Direction: boolean; RotatedType: TRotatedType): TRotatedType;
    function GetScore: Integer;
  public
    ParentField: PField;
    property CanGoDown: boolean read GetCanGoDown;
    property CanGoToLeft: boolean read GetCanGoToLeft;
    property CanGoToRight: boolean read GetCanGoToRight;
    property Canvas: TCanvas read FCanvas write FCanvas;
    property Color: TColor read FColor write SetColor;
    property Empty: boolean read GetEmpty;
    property BorderColor: TColor read FBorderColor write SetBorderColor;
    property Left: Integer read FLeft write FLeft;
    property Width: Integer read FWidth write FWidth;
    property Height: Integer read FHeight write FHeight;
    property Top: Integer read FTop write FTop;
    property FigureObjectType: TFigureObjectType read FFigureObjectType write SetFigureObjectType;
    property FigureRotated: TRotatedType read FFigureRotated write SetFigureRotated;
    property Score: Integer read GetScore;
    function CanRotate(Direction: boolean): boolean;
    function CanPut: boolean;
    procedure DoStepDown;
    procedure DoStepLeft;
    procedure DoStepRight;
    procedure DoRotate(Direction: boolean);
    procedure Draw;
    procedure ReduceLine(Line: integer);
    constructor Create;
    destructor Destroy; override;
  end;


implementation

{ TFigureObject }

constructor TFigureObject.Create;
begin
  inherited;
  FLeft := 0;
  FTop := 0;
  FWidth := 0;
  FHeight := 0;
  FColor := clBtnFace;
  ParentField := Nil;
  FBorderColor := clBtnShadow;
  FFigureObjectType := None;
  FCanvas := Nil;
  FCanGoDown := false;
  FFigureRotated := Rot0;
  Modified := false;
end;

destructor TFigureObject.Destroy;
begin
  inherited;
end;

procedure TFigureObject.DoRotate(Direction: boolean);
begin
  WriteIntoGrid(false);
  FFigureRotated := GetNextRotatedDirection(Direction, FFigureRotated);
  SetFigureType(FFigureObjectType, FFigureRotated);
  WriteIntoGrid(true);
end;

procedure TFigureObject.DoStepDown;
begin
  WriteIntoGrid(false);
  Inc(FTop);
  WriteIntoGrid(true);
end;

procedure TFigureObject.DoStepLeft;
begin
  WriteIntoGrid(false);
  Dec(FLeft);
  WriteIntoGrid(true);
end;

procedure TFigureObject.DoStepRight;
begin
  WriteIntoGrid(false);
  Inc(FLeft);
  WriteIntoGrid(true);  
end;

procedure TFigureObject.Draw;
var
  i, j: integer;
begin
  FCanvas.Brush.Color := Color;
  FCanvas.Pen.Color := BorderColor;
  for j := 1 to FHeight do
    for i := 1 to FWidth do
      if FParts[i, j] then
      begin
        FCanvas.FillRect(Rect(BarWidth * (Left + i - 2) + FieldOffsetX,
                              BarHeight * (Top + j - 2) + FieldOffsetY,
                              BarWidth * (Left + i - 1) + FieldOffsetX,
                              BarHeight * (Top + j - 1) + FieldOffsetY));
        FCanvas.Rectangle(Rect(BarWidth * (Left + i - 2) + FieldOffsetX,
                              BarHeight * (Top + j - 2) + FieldOffsetY,
                              BarWidth * (Left + i - 1) + FieldOffsetX,
                              BarHeight * (Top + j - 1) + FieldOffsetY));
      end;
end;

function TFigureObject.GetCanGoDown: boolean;
var
  i, j: integer;
begin
  Result := not Modified;
  if Result then
  begin
    Result := FTop + FHeight <= FieldHeight;
    if Result then
      for i := 1 to FWidth do
        for j := FHeight downto 1 do
          if FParts[i, j] then
          begin
            if ParentField^[FLeft + i - 1, FTop + j] then
            begin
              Result := false;
              exit;
            end
            else
              break;
          end;
  end;          
end;

function TFigureObject.GetCanGoToLeft: boolean;
var
  i, j: integer;
begin
  Result := not Modified;
  if Result then
  begin
    Result := FLeft > 1;
    if Result then
      for j := 1 to FHeight do
        for i := 1 to FWidth do
        begin
          if FParts[i, j] then
            if ParentField^[FLeft + i - 2, FTop + j - 1] then
            begin
              Result := false;
              exit;
            end
            else
              break;
        end;
  end;        
end;

function TFigureObject.GetCanGoToRight: boolean;
var
  i, j: integer;
begin
  Result := not Modified;
  if Result then
  begin
    Result := (FLeft + FWidth) <= FieldWidth;
    if Result then
      for j := 1 to FHeight do
        for i := FWidth downto 1 do
        begin
          if FParts[i, j] then
            if ParentField^[FLeft + i, FTop + j - 1] then
            begin
              Result := false;
              exit;
            end
            else
              break;
        end;
  end;        
end;

function TFigureObject.CanRotate(Direction: boolean): boolean;
var
  ParentParts: TParts;
  TempParts: TParts;
  TempRotatedType: TRotatedType;
  TempWidth, TempHeight:  integer;
  i, j: integer;
begin
  Result := not Modified;
  if Result then
  begin
    TempRotatedType := GetNextRotatedDirection(Direction, FFigureRotated);
    SetRotatedProperties(FFigureObjectType, TempRotatedType, TempParts, TempWidth, TempHeight);
    Result := ((FLeft + TempWidth - 1) <= FieldWidth) and
              ((FTop + TempHeight - 1) <= FieldHeight);
    if Result then
    begin
      for j := 1 to 4 do
        for i := 1 to 4 do
          ParentParts[i, j] := ParentField^[FLeft + i - 1, FTop + j - 1];
      for j := 1 to 4 do
        for i := 1 to 4 do
          if FParts[i, j] then
            ParentParts[i, j] := false;
      for j := 1 to 4 do
        for i := 1 to 4 do
          if ParentParts[i, j] and TempParts[i, j] then
          begin
            Result := false;
            exit;
          end;
    end;
  end;
end;

function TFigureObject.GetEmpty: boolean;
var
  i, j: integer;
begin
  for j := 1 to FHeight do
    for i := 1 to FWidth do
      if FParts[i, j] then
      begin
        Result := false;
        exit;
      end;
  Result := true;
end;

procedure TFigureObject.ReduceLine(Line: integer);
var
  i, j: integer;
begin
  WriteIntoGrid(false);

  for j := Line - FTop + 1 to FHeight - 1 do
    for i := 1 to FWidth do
      FParts[i, j] := FParts[i, j + 1];

  for i := 1 to FWidth do
    FParts[i, FHeight] := false;

  Dec(FHeight);
  WriteIntoGrid(true);
  Modified := true;
end;

procedure TFigureObject.SetBorderColor(const Value: TColor);
begin
  FBorderColor := Value;
end;

procedure TFigureObject.SetColor(const Value: TColor);
begin
  FColor := Value;
end;

procedure TFigureObject.SetFigureObjectType(const Value: TFigureObjectType);
begin
  if Value <> FFigureObjectType then
  begin
    FFigureObjectType := Value;
    SetFigureType(FFigureObjectType, Rot0);
  end;
end;

procedure TFigureObject.SetFigureRotated(const Value: TRotatedType);
begin
  if FFigureRotated <> Value then
  begin
    FFigureRotated := Value;
    SetFigureType(FFigureObjectType, FFigureRotated);
  end;
end;

procedure TFigureObject.SetRotatedProperties(RValue: TFigureObjectType; RRotated: TRotatedType;
  var RParts: TParts; var RWidth, RHeight: integer);
begin
  FillChar(RParts, SizeOf(RParts), false);
  case RValue of
    Triada:
    begin
      case RRotated of
        Rot0:
        begin
          RParts[2, 1] := true; //  x
          RParts[1, 2] := true; // xxx
          RParts[2, 2] := true;
          RParts[3, 2] := true;
          RWidth := 3;
          RHeight := 2;
        end;
        Rot90:
        begin
          RParts[1, 1] := true; // x
          RParts[1, 2] := true; // xx
          RParts[2, 2] := true; // x
          RParts[1, 3] := true;
          RWidth := 2;
          RHeight := 3;
        end;
        Rot180:
        begin
          RParts[1, 1] := true; // xxx
          RParts[2, 1] := true; //  x
          RParts[3, 1] := true;
          RParts[2, 2] := true;
          RWidth := 3;
          RHeight := 2;
        end;
        Rot270:
        begin
          RParts[2, 1] := true; //  x
          RParts[1, 2] := true; // xx
          RParts[2, 2] := true; //  x
          RParts[2, 3] := true;
          RWidth := 2;
          RHeight := 3;
        end;
      end;
    end;
    LeftCorner:
    begin
      case RRotated of
        Rot0:
        begin
          RParts[1, 1] := true; // xxx
          RParts[2, 1] := true; // x
          RParts[3, 1] := true;
          RParts[1, 2] := true;
          RWidth := 3;
          RHeight := 2;
        end;
        Rot90:
        begin
          RParts[1, 1] := true; // xx
          RParts[2, 1] := true; //  x
          RParts[2, 2] := true; //  x
          RParts[2, 3] := true;
          RWidth := 2;
          RHeight := 3;
        end;
        Rot180:
        begin
          RParts[3, 1] := true; //   x
          RParts[1, 2] := true; // xxx
          RParts[2, 2] := true;
          RParts[3, 2] := true;
          RWidth := 3;
          RHeight := 2;
        end;
        Rot270:
        begin
          RParts[1, 1] := true; // x
          RParts[1, 2] := true; // x
          RParts[1, 3] := true; // xx
          RParts[2, 3] := true;
          RWidth := 2;
          RHeight := 3;
        end;
      end;
    end;
    RightCorner:
    begin
      case RRotated of
        Rot0:
        begin
          RParts[1, 1] := true;  //  xxx
          RParts[2, 1] := true;  //    x
          RParts[3, 1] := true;
          RParts[3, 2] := true;
          RWidth := 3;
          RHeight := 2;
        end;
        Rot90:
        begin
          RParts[2, 1] := true;  //  x
          RParts[2, 2] := true;  //  x
          RParts[1, 3] := true;  // xx
          RParts[2, 3] := true;
          RWidth := 2;
          RHeight := 3;
        end;
        Rot180:
        begin
          RParts[1, 1] := true;  // x
          RParts[1, 2] := true;  // xxx
          RParts[2, 2] := true;
          RParts[3, 2] := true;
          RWidth := 3;
          RHeight := 2;
        end;
        Rot270:
        begin
          RParts[1, 1] := true; // xx
          RParts[2, 1] := true; // x
          RParts[1, 2] := true; // x
          RParts[1, 3] := true;
          RWidth := 2;
          RHeight := 3;
        end;
      end;
    end;
    LeftZikzak:
    begin
      case RRotated of
        Rot0, Rot180:
        begin
          RParts[1, 1] := true; // xx
          RParts[2, 1] := true; //  xx
          RParts[2, 2] := true;
          RParts[3, 2] := true;
          RWidth := 3;
          RHeight := 2;
        end;
        Rot90, Rot270:
        begin
          RParts[2, 1] := true; //  x
          RParts[1, 2] := true; // xx
          RParts[2, 2] := true; // x
          RParts[1, 3] := true;
          RWidth := 2;
          RHeight := 3;
        end;
      end;
    end;
    RightZikzak:
    begin
      case RRotated of
        Rot0, Rot180:
        begin
          RParts[2, 1] := true; //  xx
          RParts[3, 1] := true; // xx
          RParts[1, 2] := true;
          RParts[2, 2] := true;
          RWidth := 3;
          RHeight := 2;
        end;
        Rot90, Rot270:
        begin
          RParts[1, 1] := true; // x
          RParts[1, 2] := true; // xx
          RParts[2, 2] := true; //  x
          RParts[2, 3] := true;
          RWidth := 2;
          RHeight := 3;
        end;
      end;
    end;
    Stick:
    begin
      case RRotated of
        Rot0, Rot180:
        begin
          RParts[1, 1] := true;
          RParts[2, 1] := true;
          RParts[3, 1] := true;
          RParts[4, 1] := true;
          RWidth := 4;
          RHeight := 1;
        end;
        Rot90, Rot270:
        begin
          RParts[1, 1] := true;
          RParts[1, 2] := true;
          RParts[1, 3] := true;
          RParts[1, 4] := true;
          RWidth := 1;
          RHeight := 4;
        end;
      end;
    end;
    Box:
    begin
      RParts[1, 1] := true;
      RParts[2, 1] := true;
      RParts[1, 2] := true;
      RParts[2, 2] := true;
      RWidth := 2;
      RHeight := 2;
    end;
  end;
end;

procedure TFigureObject.SetFigureType(Value: TFigureObjectType; Rotated: TRotatedType);
begin
  SetRotatedProperties(Value, Rotated, FParts, FWidth, FHeight);
end;

procedure TFigureObject.WriteIntoGrid(DoMark: boolean);
var
  i, j: integer;
begin
  for j := 1 to FHeight do
    for i := 1 to FWidth do
      if FParts[i, j] then
        ParentField^[FLeft + i - 1, FTop + j - 1] := DoMark;
end;

function TFigureObject.GetNextRotatedDirection(Direction: boolean; RotatedType: TRotatedType): TRotatedType;
begin
  if Direction then
  begin
    if RotatedType = Rot270 then
      Result := Rot0
    else
      Result := TRotatedType(integer(RotatedType) + 1);
  end
  else
  begin
    if RotatedType = Rot0 then
      Result := Rot270
    else
      Result := TRotatedType(integer(RotatedType) - 1);
  end;
end;

function TFigureObject.CanPut: boolean;
var
  i, j: integer;
begin
  for j := 1 to FHeight do
    for i := 1 to FWidth do
      if FParts[i, j] and ParentField^[FLeft + i - 1, FTop + j - 1] then
      begin
        Result := false;
        exit;
      end;
  Result := true;
end;

function TFigureObject.GetScore: Integer;
begin
  if FFigureObjectType = None then
    Result := 0
  else
    Result := FigureScore[integer(FFigureObjectType)];
end;

end.
